# Alan Derk - 210D Final - kMC simulation of ZGB model

import numpy as np
import matplotlib.pyplot as plt

#some constants
#reactions: B{u,r} is B2 ads on the site and the {upper,right} site
#Aa is A ads on the site
#R{u,d,l,r} is reaction of A (on-site) onto B when B is {up,down,left,right}
Bad, Bar, Aa, Ru, Rd, Rl, Rr = range(7)
s, A, B = range(3)

def InitializeLattice(size=3):
    #creates a lattice with empty surface sites "s"
    lattice = np.zeros((size,size),dtype='int')
    return lattice

def PBC(size, i):
    # Applies periodic boundary conditions    
    #has i wrap around if too big -or- less than zero

    if i < 0:
        i += size
    if i > size - 1:
        i -= size
    return i

def PBC_tuple(size,i):
    #Applies PBC to a tuple
    return tuple([PBC(size,j) for j in i])
        
def GetRxnsAtSite(lattice,site) :
    #returns all possible sites, rxns for a single site
    site = np.array(site)    
    sites = []
    rxns = []
    size = len(lattice)
    #print "GetRxnsAtSite: size={0} site={1} type={2}".format(size,site,lattice[tuple(site)])        
    if lattice[tuple(site)] == s:
        #empty site found
        sites.append(site)
        rxns.append(Aa)
        #print "size={0} site={1}".format(size,site)        
        if lattice[PBC_tuple(size, site + (1,0))] == s:
            sites.append(site)
            rxns.append(Bad)
        if lattice[PBC_tuple(size, site + (0,1))] == s:
            sites.append(site)
            rxns.append(Bar)
    elif lattice[tuple(site)] == A:
        #Found adsorbed A, see if any B around to react with
        if lattice[PBC_tuple(size, site + (1,0))] == B:
            sites.append(site)
            rxns.append(Rd)
        if lattice[PBC_tuple(size, site + (-1,0))] == B:
            sites.append(site)
            rxns.append(Ru)
        if lattice[PBC_tuple(size, site + (0,1))] == B:
            sites.append(site)
            rxns.append(Rr)
        if lattice[PBC_tuple(size, site + (0,-1))] == B:
            sites.append(site)
            rxns.append(Rl)
    #print "at site={0} type={2} returning sites={1}".format(site,sites)
    return sites, rxns
    
def GenerateProcessList(lattice):
    size = len(lattice)    
    tot_sites = []
    tot_rxns = []
    for i in range(size):
        for j in range(size):
            sites, rxns = GetRxnsAtSite(lattice,(i,j))   
            tot_sites += sites
            tot_rxns += rxns
    return np.array(tot_sites), np.array(tot_rxns)
    

def CalcWba(rxns,y):
    # Returns list of transition probabilities for each rxn in rxns
    Wba = np.zeros(len(rxns))
    for rxn in (Bad, Bar, Aa, Ru, Rd, Rl, Rr):
        if rxn == Bad or rxn == Bar:
            Wba[rxns == rxn] = 1 - y
        elif rxn == Aa:
            Wba[rxns == rxn] = y
        else:
            Wba[rxns == rxn] = 50
    return Wba
    
def UpdateProcessList(site, lattice, sites, rxns):
    # Updates sites and rxns to reflect a change at site and neighbors
    # returns sites, rxns
    size = len(lattice)
    neighbors = [(0,0),(1,0),(0,1),(-1,0),(0,-1)]
    for step in neighbors:
        #print "step={0} site={1}".format(step,site)
        loc = PBC_tuple(size, step + np.array(site))
        print "UpdateProcessList: Deleting loc={0} no_rxns={1}".format(loc,len(rxns))
        keepers = np.apply_along_axis(np.any,1,sites!=loc)
        #removes all processes at loc
        #print "keepers:{}".format(keepers)        
        #print "sites:{}".format(sites)        
        sites = sites[keepers]
        #print "sites after removal:{}".format(sites)        
        rxns = rxns[keepers]
    for step in neighbors:
        #add back in new reactions
        loc = PBC_tuple(size, step + np.array(site))
        newsites, newrxns = GetRxnsAtSite(lattice,loc)
        print "UpdateProcessList: Creating loc={0} no_rxns={1}".format(loc,len(rxns))
        if newsites != []:           
            sites = np.concatenate((sites,newsites))
            rxns = np.concatenate((rxns,newrxns))    
    return sites, rxns
        
        
    
        
#Parameters
#Mole fraction of A in gas            
y = .9
#End time
tmax = 50
size=24


lattice = InitializeLattice(size=size)
fig = plt.figure()
ax = fig.add_subplot(111)
ax.imshow(lattice, interpolation='nearest')

t = 0
amt_C = 0
i = 0
while t <= tmax:
    sites, rxns = GenerateProcessList(lattice)
    Wba = CalcWba(rxns,y)    
    Ka = np.sum(Wba)    # probability of transitioning out of current state
    t += -np.log(np.random.random()) / Ka #increase time
    event = np.random.choice(range(len(rxns)), p=Wba/Ka)
    rxn = rxns[event]
    site = sites[event][0]
    #print "\nMC step rxn={0} site={1} no_rxns={2}".format(rxn, site,len(rxns))
    #print "lattice="
    #print lattice
    #Apply event and update process list
    if rxn == Bad:
        lattice[PBC_tuple(size, site)] = B
        neighbor = PBC_tuple(size, site+ (1,0))        
        lattice[neighbor] = B
    elif rxn == Bar:
        lattice[PBC_tuple(size, site)] = B
        neighbor = PBC_tuple(size, site+ (0,1))        
        lattice[neighbor] = B
    elif rxn == Aa:
        lattice[PBC_tuple(size, site)] = A
    elif rxn == Rd:
        lattice[PBC_tuple(size, site)] = s
        neighbor = PBC_tuple(size, site+ (1,0))
        lattice[neighbor] = s
    elif rxn == Rl:
        lattice[PBC_tuple(size, site)] = s
        neighbor = PBC_tuple(size, site+ (0,-1))
        lattice[neighbor] = s
    elif rxn == Rr:
        lattice[PBC_tuple(size, site)] = s
        neighbor = PBC_tuple(size, site+(0,1))
        lattice[neighbor] = s
    elif rxn == Ru:
        lattice[PBC_tuple(size, site)] = s
        neighbor = PBC_tuple(size, site+(-1,0))        
        lattice[neighbor] = s
    if rxn == Rd or rxn == Rl or rxn == Rr or rxn == Ru:
        amt_C += 1        
        #print "made C! amt_C={0} rate={1}".format(amt_C,amt_C/t)
    A_cov = np.sum(lattice == A)/float(size**2)
    print "time={0} rxn_rate={1} coverage={2}".format(t,amt_C/t,A_cov)
    ax.imshow(lattice, interpolation='nearest')
    i += 1
    plt.savefig("mov{0:04}.png".format(i),dpi=96,padding=0)

    
    

    
    
    
    
         
    
    
        
        
        
        
        
        
        
        
    
    
